/*
 * bonobo-stream-efs.c: libefs based Stream implementation
 *
 * Author:
 *   Dietmar Maurer (dm@vlsivie.tuwien.ac.at)
 *
 */

#include <config.h>
#include <storage-modules/bonobo-stream-efs.h>

static BonoboStream *
create_stream_efs_server (const BonoboStreamEFS *stream_efs)
{
	Bonobo_Stream corba_stream;

	corba_stream = bonobo_stream_corba_object_create (
		BONOBO_OBJECT (stream_efs));

	return BONOBO_STREAM (
		bonobo_object_construct (
			BONOBO_OBJECT (stream_efs), 
			corba_stream));
}

static void
bonobo_stream_efs_destroy (GtkObject *object)
{
	BonoboStreamEFS *stream_efs = BONOBO_STREAM_EFS (object);

	if (stream_efs->file)
		efs_file_close (stream_efs->file);

	if (stream_efs->storage) 
		bonobo_object_unref (BONOBO_OBJECT (stream_efs->storage));    
}

static CORBA_long
real_write (BonoboStream *stream, const Bonobo_Stream_iobuf *buffer,
	    CORBA_Environment *ev)
{
	BonoboStreamEFS *stream_efs = BONOBO_STREAM_EFS (stream);

	if (efs_file_write (stream_efs->file, buffer->_buffer, 
			    buffer->_length)) {
		CORBA_exception_set (ev, CORBA_USER_EXCEPTION, 
				     ex_Bonobo_Stream_IOError, NULL);
		return 0;
	}

	return buffer->_length;
}

static void
real_read (BonoboStream *stream, CORBA_long count,
	   Bonobo_Stream_iobuf ** buffer,
	   CORBA_Environment *ev)
{
	BonoboStreamEFS *stream_efs = BONOBO_STREAM_EFS (stream);
	CORBA_octet *data;
	gint32 bytes_read;
	EFSResult result;

	if (count < 0) {
		CORBA_exception_set (ev, CORBA_USER_EXCEPTION, 
				     ex_Bonobo_Stream_IOError, NULL);
		return;
	}

	*buffer = Bonobo_Stream_iobuf__alloc ();
	CORBA_sequence_set_release (*buffer, TRUE);
	data = CORBA_sequence_CORBA_octet_allocbuf (count);

	result = efs_file_read (stream_efs->file, data, count, &bytes_read);
	
	if (!result || (result == EFS_ERR_EOF)) {
		(*buffer)->_buffer = data;
		(*buffer)->_length = bytes_read;
	} else {
		CORBA_free (data);
		CORBA_free (*buffer);
		*buffer = NULL;
		CORBA_exception_set (ev, CORBA_USER_EXCEPTION, 
				     ex_Bonobo_Stream_IOError, NULL);
	}
}

static CORBA_long
real_seek (BonoboStream *stream, CORBA_long offset, Bonobo_Stream_SeekType whence,
	   CORBA_Environment *ev)
{
	BonoboStreamEFS *stream_efs = BONOBO_STREAM_EFS (stream);
	int fw;
	guint32 pos;

	if (whence == Bonobo_Stream_SEEK_CUR)
		fw = EFS_SEEK_CUR;
	else if (whence == Bonobo_Stream_SEEK_END)
		fw = EFS_SEEK_END;
	else
		fw = EFS_SEEK_SET;
	
	if (efs_file_seek (stream_efs->file, offset, fw, &pos)) {
		CORBA_exception_set (ev, CORBA_USER_EXCEPTION, 
				     ex_Bonobo_Stream_IOError, NULL);
		return 0;
	}
	
	return pos;
}

static void
real_truncate (BonoboStream *stream, const CORBA_long new_size, 
	       CORBA_Environment *ev)
{
	BonoboStreamEFS *stream_efs = BONOBO_STREAM_EFS (stream);

	if (efs_file_trunc (stream_efs->file, new_size) < 0)
		CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
				     ex_Bonobo_Stream_NoPermission, NULL);
}

static void
real_close (BonoboStream      *stream,
	    CORBA_Environment *ev)
{
	BonoboStreamEFS *stream_efs = BONOBO_STREAM_EFS (stream);

	if (efs_file_close (stream_efs->file) < 0)
		CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
				     ex_Bonobo_Stream_IOError, NULL);
	stream_efs->file = NULL;
}

static void
bonobo_stream_efs_class_init (BonoboStreamEFSClass *class)
{
	GtkObjectClass *object_class = (GtkObjectClass *) class;
	BonoboStreamClass *sclass = BONOBO_STREAM_CLASS (class);
	
	sclass->write    = real_write;
	sclass->read     = real_read;
	sclass->seek     = real_seek;
	sclass->truncate = real_truncate;
        sclass->close    = real_close;
	/* FIXME: impl eof ! */

	object_class->destroy = bonobo_stream_efs_destroy;
}

GtkType
bonobo_stream_efs_get_type (void)
{
	static GtkType type = 0;

	if (!type) {
		GtkTypeInfo info = {
			"IDL:GNOME/StreamEFS:1.0",
			sizeof (BonoboStreamEFS),
			sizeof (BonoboStreamEFSClass),
			(GtkClassInitFunc) bonobo_stream_efs_class_init,
			(GtkObjectInitFunc) NULL,
			NULL, /* reserved 1 */
			NULL, /* reserved 2 */
			(GtkClassInitFunc) NULL
		};

		type = gtk_type_unique (bonobo_stream_get_type (), &info);
	}
  
	return type;
}

static BonoboStream *
bonobo_stream_efs_open_create (BonoboStorageEFS *storage_efs,
			       const CORBA_char *path,
			       Bonobo_Storage_OpenMode mode,
			       gboolean create)
{
	BonoboStreamEFS *stream_efs;
	gint flags;

	g_return_val_if_fail (path != NULL, NULL);
	g_return_val_if_fail (storage_efs != NULL, NULL);

	if (!(stream_efs = gtk_type_new (bonobo_stream_efs_get_type ()))) 
		return NULL;

	flags = EFS_READ;

	if (mode & Bonobo_Storage_WRITE)
		flags |= EFS_WRITE;

	if (create)
		flags |= EFS_CREATE;
  
	if (efs_file_open (&stream_efs->file, storage_efs->dir, path, flags)) {
		bonobo_object_unref (BONOBO_OBJECT (stream_efs));
		return NULL;
	}
	
	stream_efs->storage = storage_efs;
	bonobo_object_ref (BONOBO_OBJECT(storage_efs));
	create_stream_efs_server (stream_efs);

	return BONOBO_STREAM (stream_efs);
}

BonoboStream *
bonobo_stream_efs_open (BonoboStorageEFS *storage, const CORBA_char *path, 
			Bonobo_Storage_OpenMode mode)
{
	return bonobo_stream_efs_open_create (storage, path, mode, FALSE);
}

BonoboStream *
bonobo_stream_efs_create (BonoboStorageEFS *storage, const CORBA_char *path)
{
	return bonobo_stream_efs_open_create (storage, path, 0, TRUE);
}
